#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <sys/types.h>          /* See NOTES */
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "json.h"
#include "defs.h"

static char *config_file = DEFAULT_CONFIG_FILE;
FILE *http;
int verbose;
char const http_version[] = "HTTP/1.1";

enum {
	EX_P903_ALIVE = 0,
	EX_P903_NOTALIVE = 1,
	EX_P903_MIXED = 2
};

static char *
read_listen(char **ret_service)
{
	FILE *fp;
	int ln = 0;
	char buf[1024];
	int skip_to_eol = 0;
	char *retval = NULL;

	*ret_service = NULL;

	fp = fopen(config_file, "r");
	if (!fp) {
		if (errno == ENOENT)
			return NULL;
		else {
			fatal("can't open %s: %s",
			      config_file, strerror(errno));
			exit(EX_CONFIG);
		}
	}

	while (fgets(buf, sizeof(buf), fp)) {
		int len;
		char *p;

		ln++;
		len = strlen(buf);
		if (len == 0)
			continue;

		if (skip_to_eol) {
			skip_to_eol = buf[len-1] != '\n';
			continue;
		}

		if (buf[len-1] == '\n')
			buf[--len] = 0;
		else if (!feof(fp)) {
			error("%s:%d: line too long", config_file, ln);
			skip_to_eol = 1;
			continue;
		}

		while (len > 0 && isspace(buf[len-1]))
			buf[--len] = 0;

		if (len == 0)
			continue;

		for (p = buf; (*p == ' ' || *p == '\t'); p++)
			;
		if (*p == 0 || *p == '#')
			continue;

		if (strncmp(p, "listen", 6) == 0) {
			char *q;

			p += 6;
			while (*p && (*p == ' '||*p == '\t'))
				p++;
			retval = estrdup(p);
			q = strchr(retval, ':');
			if (q) {
				*q++ = 0;
				if (*q) {
					*ret_service = estrdup(q);
				}
			}
			break;
		}
	}

	fclose(fp);
	return retval;
}

static void
http_connect(char *node, char *service)
{
	struct addrinfo hints;
	struct addrinfo *res;
	int fd;
	int rc;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	rc = getaddrinfo(node, service, &hints, &res);
	if (rc) {
		fatal("%s: %s", node, gai_strerror(rc));
		exit(EX_UNAVAILABLE);
	}

	fd = socket(res->ai_family, res->ai_socktype, 0);
	if (fd == -1) {
		fatal("socket: %s", strerror(errno));
		exit(EX_UNAVAILABLE);
	}
	if (connect(fd, (struct sockaddr *)res->ai_addr, res->ai_addrlen)) {
		fatal("failed to connect to %s:%s: %s",
		      node, service, strerror(errno));
		exit(EX_UNAVAILABLE);
	}
	http = fdopen(fd, "w+");
	if (http == NULL) {
		fatal("failed to open socket: %s", strerror(errno));
		exit(EX_UNAVAILABLE);
	}
	freeaddrinfo(res);
}

struct http_resp {
	int code;     /* HTTP response code */
	char *reason; /* Human-readable status */
	char **hv;    /* Headers */
	size_t hc;    /* Number of headers in hv */
	size_t hn;    /* Max capacity of hv */
	char *content;
	size_t content_length;
};

struct http_buf {
	char *base;
	size_t size;
	size_t len;
};

static char const *
http_resp_get_header(struct http_resp *resp, char const *name)
{
	int i;
	size_t len = strlen(name);

	for (i = 0; i < resp->hc; i++) {
		if (strlen(resp->hv[i]) > len
		    && resp->hv[i][len] == ':'
		    && strncasecmp(resp->hv[i], name, len) == 0) {
			char const *ret = resp->hv[i] + len + 1;
			while (*ret && (*ret == ' ' || *ret == '\t'))
				ret++;
			return ret;
		}
	}
	return NULL;
}

static void
http_readline(struct http_buf *hbuf)
{
	hbuf->len = 0;
	while (1) {
		if (hbuf->len == hbuf->size) {
			hbuf->base = e2nrealloc(hbuf->base, &hbuf->size, 1);
		}
		if (fgets(hbuf->base + hbuf->len, hbuf->size, http) == NULL) {
			if (feof(http)) {
				fatal("connection closed prematurely");
				exit(EX_PROTOCOL);
			} else {
				fatal("error reading from socket: %s",
				      strerror(errno));
				exit(EX_OSERR);
			}
		}
		hbuf->len += strlen(hbuf->base + hbuf->len);
		if (hbuf->base[hbuf->len-1] == '\n') {
			hbuf->base[--hbuf->len] = 0;
			if (hbuf->base[hbuf->len-1] == '\r')
				hbuf->base[--hbuf->len] = 0;
			break;
		}
	}
}

static int
split3(const char *input, char *res[3], int flag)
{
	size_t len;
	char *p;

	p = strchr(input, ' ');
	if (!p)
		return 1;

	len = p - input;
	res[0] = malloc(len + 1);
	if (!res[0])
		return -1;
	memcpy(res[0], input, len);
	res[0][len] = 0;

	input = p + 1;

	p = (flag ? strrchr : strchr)(input, ' ');

	if (!p)
		return 1;
	len = p - input;
	res[1] = malloc(len + 1);
	if (!res[1])
		return -1;
	memcpy(res[1], input, len);
	res[1][len] = 0;

	res[2] = strdup(p + 1);
	if (!res[2])
		return -1;

	return 0;
}

int
strsplit3(const char *input, char *result[3], int flag)
{
	char *tmp[3] = { NULL, NULL, NULL };
	int rc = split3(input, tmp, flag);
	if (rc)	{
		int ec = errno;
		free(tmp[0]);
		free(tmp[1]);
		free(tmp[2]);
		errno = ec;
	} else {
		result[0] = tmp[0];
		result[1] = tmp[1];
		result[2] = tmp[2];
	}
	return rc;
}

static void
http_resp_init(struct http_resp *resp)
{
	memset(resp, 0, sizeof(*resp));
}

static void
http_resp_reset(struct http_resp *resp)
{
	size_t i;

	resp->code = 0;
	free(resp->reason);
	resp->reason = NULL;
	for (i = 0; i < resp->hc; i++) {
		free(resp->hv[i]);
		resp->hv[i] = NULL;
	}
	resp->hc = 0;
	free(resp->content);
	resp->content = NULL;
	resp->content_length = 0;
}

static void
http_recv(struct http_resp *resp)
{
	enum input_state { is_initial, is_headers, is_content } state = is_initial;
	struct http_buf hbuf;
	char const *hval;

	memset(&hbuf, 0, sizeof(hbuf));
	http_resp_reset(resp);
	while (state != is_content) {
		http_readline(&hbuf);
		if (state == is_initial) {
			char *res[3];

			if (strsplit3(hbuf.base, res, 0)) {
				fatal("malformed HTTP response");
				exit(EX_PROTOCOL);
			}
			if (strcmp(res[0], http_version)) {
				fatal("unsupported HTTP version: %s", res[0]);
				exit(EX_PROTOCOL);
			}
			free(res[0]);

			resp->code = atoi(res[1]);
			if (resp->code <= 0 || resp->code > 559) {
				fatal("bad response code: %s", res[1]);
				exit(EX_PROTOCOL);
			}
			free(res[1]);
			resp->reason = res[2];
			state = is_headers;
		} else if (state == is_headers) {
			if (resp->hc == resp->hn) {
				resp->hv = json_2nrealloc(resp->hv,
							  &resp->hn,
							  sizeof(resp->hv[0]));
			}
			if (hbuf.len > 0) {
				resp->hv[resp->hc++] = estrdup(hbuf.base);
			} else {
				resp->hv[resp->hc] = NULL;
				state = is_content;
			}
		}
	}

	hval = http_resp_get_header(resp, "content-length");
	if (hval) {
		char *p;
		unsigned long len = strtoul(hval, &p, 10);
		if (*p) {
			fatal("malformed content-length");
			exit(EX_PROTOCOL);
		}
		resp->content = emalloc(len + 1);
		resp->content_length = len;

		p = resp->content;
		while (len) {
			ssize_t n = fread(p, 1, len, http);
			if (n == -1) {
				fatal("socket read error: %s",
				      strerror(errno));
				exit(EX_OSERR);
			}
			if (n == 0) {
				error("warning: short read from socket");
				exit(EX_OSERR);
			}
			len -= n;
			p += n;
		}
		*p = 0;
	}
}

static void
http_query(char const *meth, char const *url, char **hdr)
{
	size_t i;
	int host = 0;

	fprintf(http, "%s %s HTTP/1.1\r\n", meth, url);
	if (hdr) {
		for (i = 0; hdr[i]; i++) {
			fprintf(http, "%s\r\n", hdr[i]);
			if (!host && strncasecmp(hdr[i], "host:", 5) == 0)
				host = 1;
		}
	}
	if (!host)
		fprintf(http, "Host: localhost\r\n");
	fprintf(http, "\r\n");
}

static char const *
json_strerror(int ec)
{
	static char *json_errstr[] = {
		[JSON_E_NOERR]     = "No error",
		[JSON_E_NOMEM]     = "Not enough memory",
		[JSON_E_BADTOK]    = "Unrecognized token",
		[JSON_E_BADDELIM]  = "Bad delimiter",
		[JSON_E_BADSTRING] = "Malformed string"
	};
	if (ec < 0 || ec >= sizeof(json_errstr) / sizeof(json_errstr[0]))
		return "Unknown error";
	return json_errstr[ec];
}

static struct json_value *
ejson_get(struct json_value *obj, char *name, int type)
{
	struct json_value *jv;
	if (json_object_get(obj, name, &jv)) {
		fatal("no \"%s\" member in the response", name);
		exit(EX_DATAERR);
	}
	if (jv->type != type) {
		fatal("\"%s\" member has wrong type", name);
		exit(EX_DATAERR);
	}
	return jv;
}

static int
print_host_status(struct json_value *obj)
{
	struct json_value *jv;
	int alive;
	char const *name;

	jv = ejson_get(obj, "name", json_string);
	name = jv->v.s;

	jv = ejson_get(obj, "status", json_bool);
	if (!jv->v.b) {
		printf("no response from %s\n", name);
		return EX_P903_NOTALIVE;
	}

	alive = ejson_get(obj, "alive", json_bool)->v.b;
	printf("%s %s\n", name, alive ? "alive" : "not alive");
	if (verbose) {
		double xmit = ejson_get(obj, "xmit", json_number)->v.n;
		double recv = ejson_get(obj, "recv", json_number)->v.n;
		double start_ts = ejson_get(obj, "start-timestamp",
					   json_number)->v.n;
		double stop_ts = ejson_get(obj, "stop-timestamp",
					   json_number)->v.n;

		printf("--- %s ping statistics ---\n", name);
		printf("%lu packets transmitted, %lu received, %d%% packet loss, time %.0fms\n",
		       (unsigned long) xmit, (unsigned long) recv,
		       (int)(100 * (xmit - recv) / recv),
		       (stop_ts - start_ts) * 1000);
		printf("rtt min/avg/max/mdev = %.3f/%.3f/%.3f/%.3f ms\n",
		       ejson_get(obj, "tmin", json_number)->v.n,
		       ejson_get(obj, "avg", json_number)->v.n,
		       ejson_get(obj, "tmax", json_number)->v.n,
		       ejson_get(obj, "stddev", json_number)->v.n);
	}
	return alive ? EX_P903_ALIVE : EX_P903_NOTALIVE;
}

static void
query_host(char const *name)
{
	int rc;
	struct http_resp resp;
	struct json_value *obj;
	char url[1024];
	char const *hval;
	char *p;
	ssize_t n = snprintf(url, sizeof(url), "/host/%s", name);
	if (n < 0 || n == sizeof(url)) {
		fatal("bad host name or IP");
		exit(EX_USAGE);
	}
	http_query("GET", url, NULL);
	http_resp_init(&resp);
	http_recv(&resp);
	if (resp.code != 200) {
		fatal("%s", resp.reason);
		exit(EX_UNAVAILABLE);
	}

	hval = http_resp_get_header(&resp, "content-type");
	if (!hval || strcmp(hval, "application/json")) {
		fatal("missing or unsupported content type");
		exit(EX_UNAVAILABLE);
	}

	rc = json_parse_string(resp.content, &obj, &p);
	if (rc != JSON_E_NOERR) {
		fatal("%s near %s", json_strerror(rc), p);
		exit(EX_DATAERR);
	}
	if (obj->type != json_object) {
		fatal("returned entity has wrong type");
		exit(EX_DATAERR);
	}
	exit(print_host_status(obj));
}

static void
query_all(void)
{
	int rc;
	struct http_resp resp;
	struct json_value *obj;
	char const *hval;
	char *p;
	size_t count[2] = { 0, 0 };
	size_t len;
	size_t i;

	http_query("GET", "/host", NULL);
	http_resp_init(&resp);
	http_recv(&resp);
	if (resp.code != 200) {
		fatal("%s", resp.reason);
		exit(EX_UNAVAILABLE);
	}

	hval = http_resp_get_header(&resp, "content-type");
	if (!hval || strcmp(hval, "application/json")) {
		fatal("missing or unsupported content type");
		exit(EX_UNAVAILABLE);
	}

	rc = json_parse_string(resp.content, &obj, &p);
	if (rc != JSON_E_NOERR) {
		fatal("%s near %s", json_strerror(rc), p);
		exit(EX_DATAERR);
	}
	if (obj->type != json_array) {
		fatal("returned entity has wrong type");
		exit(EX_DATAERR);
	}
	len = json_array_length(obj);
	for (i = 0; i < len; i++) {
		struct json_value *jv;
		if (json_array_get(obj, i, &jv)) {
			fatal("can't get element %lu", (unsigned long) i);
			exit(EX_SOFTWARE);
		}
		if (jv->type == json_object)
			count[print_host_status(jv)]++;
	}
	if (count[1] == 0)
		exit(EX_P903_ALIVE);
	if (count[0] == 0)
		exit(EX_P903_NOTALIVE);
	exit(EX_P903_MIXED);
}

void
usage(void)
{
	printf("Usage: %s [-hVv] [-c FILE]\n", progname);
	printf("Query ping903 daemon.\n");
	printf("\n");
	printf("Options:\n\n");
	printf("  -c FILE        read configuration from FILE\n");
	printf("  -h             print this help test\n");
	printf("  -V             print program version and exit\n");
	printf("  -v             additional verbosity\n");
	printf("\n");
	printf("Report bugs to <%s>.\n", PACKAGE_BUGREPORT);
	printf("%s home page: <%s>\n", PACKAGE_NAME, PACKAGE_URL);
}

void
version(void)
{
	printf("%s (%s) %s\n", progname, PACKAGE_NAME, PACKAGE_VERSION);
	printf("%s", COPYLEFT);
}

int
main(int argc, char **argv)
{
	int c;
	char *p, *node, *service;

	set_progname(argv[0]);
	if (argc == 2) {
		if (strcmp(argv[1], "--help") == 0) {
			usage();
			exit(0);
		}
		if (strcmp(argv[1], "--version") == 0) {
			version();
			exit(0);
		}
	}
	while ((c = getopt(argc, argv, "c:hVv")) != EOF) {
		switch (c) {
		case 'c':
			config_file = optarg;
			break;
		case 'h':
			usage();
			exit(0);
		case 'V':
			version();
			exit(0);
		case 'v':
			verbose++;
			break;
		default:
			exit(EX_USAGE);
		}
	}
	argc -= optind;
	argv += optind;

	p = node = read_listen(&service);
	if (!node || !node[0])
		node = "127.0.0.1";
	if (!service)
		service = DEFAULT_SERVICE;
	http_connect(node, service);
	free(p);

	switch (argc) {
	case 0:
		query_all();
	case 1:
		query_host(argv[0]);
	default:
		fatal("bad number of arguments");
	}
	exit(EX_USAGE);
}
